;
; ABSTRACO GRFICA
; Uma caneta sobre uma folha de papel;
;
; F. Nunes Ferreira
; fnf@fe.up.pt
; 30 de Agosto de 2001
;


; Incluso da biblioteca grfica da verso 53 do DrScheme
;

(require-library "graphics.ss" "graphics")

;
; No DrScheme, Graphical Full Scheme (MrEd)
; a constante pi no est definida. Sendo assim:
;
(require-library "varios.scm" "user-feup")

; Primitivas grficas disponveis
; na abstrao Caneta sobre uma folha de papel
; --------------------------------------------
;
; Trata-se de uma abstraco, baseada numa janela
; grfica em que uma caneta desenha e pinta pontos, rectas 
; e elipses, e tambm desenha texto.
;
; As primitivas grficas
; e respectivos parmetros so as seguintes:
;
; janela: largura, altura, titulo
;    Cria uma janela grfica.
;
; cor: indice (inteiro entre 0 e 26)
;    Para seleccionar a cor da caneta.
;
; move: ponto (lista composta por x y)
;    Sem desenhar, move a caneta para o ponto.
;
; move-rel: ponto definido relativo  posio da 
;           caneta (lista composta por dx dy)
;    Sem desenhar, move a caneta para o ponto.
;
; desenha: pontos (lista de pontos)
;    Desenha a poligonal, definida por uma 
;    sequncia de pontos.
;    No final, a caneta ficar sobre o ltimo 
;    ponto da sequncia.
;
; desenha-oval: dois-pontos (lista de dois pontos: o
;               vrtices superior-esquerdo e 
;               inferior-direito de um rectngulo).
;    Desenha oval inscrita no rectngulo.
;    O rectngulo  definido por uma lista de 2
;    vrtices, especificados pela seguinte ordem: 
;    superior-esquerdo e inferior-direiro
;    No final, a caneta mantm a posio de partida.
;
; pinta: pontos (lista de pontos)
;    Pinta o polgono, definido por uma 
;    sequncia de pontos.
;    No final, a caneta ficar sobre o ltimo 
;    ponto da sequncia.
;
; pinta-oval: dois-pontos (lista de dois pontos: o
;             vrtices superior-esquerdo e 
;             inferior-direito de um rectngulo).
;    Pinta oval inscrita no rectngulo.
;    O rectngulo  definido por uma lista de 2
;    vrtices, especificados pela seguinte ordem: 
;    superior-esquerdo e inferior-direiro.
;    No final, a caneta mantm a posio de partida.
; 
; ponto: ponto (lista composta por x y)
;    Pinta o ponto e, no final, a caneta ficar sobre 
;    esse ponto.
;
; desenha-rel: pontos (lista de pontos relativos  posio da
;              caneta. Como a caneta se vai deslocando, cada 
;              ponto, excepto o primeiro,  definido 
;              relativamente ao ponto anterior.
;    Desenha a poligonal, definida pela 
;    sequncia de pontos.
;    No final, a caneta ficar sobre o ltimo 
;    ponto da sequncia.
;
; desenha-oval-rel: dois-pontos (lista de dois pontos, ambos
;                   definidos relativamente  posio da 
;                   caneta: o vrtice superior-esquerdo e 
;                   inferior-direito de um rectngulo).
;       Desenha oval inscrita no rectngulo.
;       O rectngulo  definido por uma lista de 2
;       vrtices, especificados pela seguinte ordem: 
;       superior-esquerdo e inferior-direiro.
;       No final, a caneta retoma a posio 
;       que tinha  partida. 
;
; pinta-rel:  pontos (lista de pontos relativos  posio da
;             caneta. Como a caneta se vai deslocando, cada 
;             ponto, excepto o primeiro,  definido 
;             relativamente ao ponto anterior.
;    Pinta o polgono, definido pela sequncia de pontos.
;    No final, a caneta ficar sobre o ltimo ponto.
;
; ponto-rel: ponto definido relativo  posio da 
;            caneta (lista composta por dx dy)
;    Desenha ponto e deixa a caneta nesse ponto.
;
; desenha-txt: texto (cadeia de caracteres)
;    Desenha o texto especificado, a partir 
;    da posio da caneta.
;    A caneta fica na parte final do texto.
;
; caneta: no tem parmetros
;    Devolve a posio actual da caneta.
;
; cursor: no tem parmetros
;    Devolve o ponto do interior da janela
;    apontado pelo cursor, quando  actuada 
;    o boto da esquerda
;
; limpa: no tem parmetros
;    Limpa a janela grfica.
;

(define graf-posi (list 0 0))                ; posicao inicial da caneta
(define graf-cor-caneta (make-rgb 0 0 0))    ; cor inicial da caneta
(define graf-cor-temp? #f)                   ; cor temporaria?
(define graf-vp 0)                           ; viewport
(define graf-altura 0)                       ; altura do viewport
;
; cria uma janela grfica
;
(define janela               
  (lambda (larg alt titulo)
    (open-graphics)
    (set! graf-altura alt)
    (let ((vport (open-viewport titulo larg alt)))
      (set! graf-vp vport))))
;
;  para seleccionar a cor da caneta ...
;
(define cor      
  (lambda (indice)       ; indice = 'temp (passando duas vezes com esta cor
    ;                    ; repe a cor existente antes da 1 passagem
    (let ((tab-cores
	   '((0.0  0.0  0.0) (0.0  0.0  0.5) (0.0  0.0  1.0) ; 0,1,3
             (0.0  0.5  0.0) (0.0  0.5  0.5) (0.0  0.5  1.0) ; 3,4,5
             (0.0  1.0  0.0) (0.0  1.0  0.5) (0.0  1.0  1.0) ; 6,7,8
             (0.5  0.0  0.0) (0.5  0.0  0.5) (0.5  0.0  1.0) ; 9,10,11
             (0.5  0.5  0.0) (0.5  0.5  0.5) (0.5  0.5  1.0) ; 12,13,14
             (0.5  1.0  0.0) (0.5  1.0  0.5) (0.5  1.0  1.0) ; 15,16,17
             (1.0  0.0  0.0) (1.0  0.0  0.5) (1.0  0.0  1.0) ; 18,19,20
             (1.0  0.5  0.0) (1.0  0.5  0.5) (1.0  0.5  1.0) ; 21,22,23
             (1.0  1.0  0.0) (1.0  1.0  0.5) (1.0  1.0  1.0)))) ; 24,25,26
      (cond ((equal? indice 'temp)
             (set! graf-cor-temp? #t))
            (else
	     (let ((cor-aux (list-ref tab-cores indice)))
	       (let ((cor-drscm (make-rgb (car cor-aux)
					  (cadr cor-aux)
					  (caddr cor-aux))))
		 (set! graf-cor-temp? #f)
		 (set! graf-cor-caneta cor-drscm))))))))
;
; Sem desenhar, move a caneta para um ponto
;
(define move             
  (lambda (ponto)        
    (set-car! graf-posi (car ponto))
    (set-car! (cdr graf-posi) (cadr ponto))))
;
; Sem desenhar, move a caneta para um ponto, o qual 
; definido relativo  posio corrente da caneta
;
(define move-rel        
  (lambda (d)            ; o parmetro d  a lista (dx dy)
    (set-car! graf-posi (+ (car graf-posi)
                           (car d)))
    (set-car! (cdr graf-posi) (+ (cadr graf-posi)
                                 (cadr d)))))
;
; desenha uma poligonal, definida por uma sequncia de pontos
;
;      No final, a caneta ficar sobre o ltimo ponto da sequncia.
;
(define desenha         
  (lambda (seq-pontos)  
    (cond ((and (> (length seq-pontos) 1)
                graf-cor-temp?)
           ((flip-line graf-vp)       ; cor temporria
            (pt->posn (car seq-pontos) '(0 0))
            (pt->posn (cadr seq-pontos) '(0 0)))
           (desenha (cdr seq-pontos)))
          ((> (length seq-pontos) 1)
           ((draw-line graf-vp)       ; cor escolhida
            (pt->posn (car seq-pontos) '(0 0))
            (pt->posn (cadr seq-pontos) '(0 0))
            graf-cor-caneta)
           (desenha (cdr seq-pontos))))
    ;                                      
    ;  e deixa caneta no ltimo ponto
    ;
    (let ((ultimo (car (reverse seq-pontos))))  
      (set-car! graf-posi (car ultimo))
      (set-car! (cdr graf-posi) (cadr ultimo)))))
;
; desenha oval inscrita num rectngulo
;
; O rectngulo  definido por uma lista de 2 vrtices,
; especificados pela seguinte ordem: 
;             superior-esquerdo e inferior-direiro
;
;         No final, a caneta mantm a posio de partida.
;
(define desenha-oval          
  (lambda (dois-vertices)  
    (let* ((v1 (car dois-vertices))
	   (v2 (cadr dois-vertices))
	   (v1x (car v1))
	   (v1y (cadr v1))
	   (larg (- (car v2) (car v1)))
	   (alt (- (cadr v1) (cadr v2))))
      (if graf-cor-temp?
          ((flip-ellipse graf-vp) (pt->posn v1 '(0 0)) larg alt)  
          
          ((draw-ellipse graf-vp) (pt->posn v1 '(0 0)) larg alt graf-cor-caneta)))))    
;
; pinta um polgono, definida por uma sequncia de pontos.
; Se o ltimo ponto for diferente do primeiro, o polgono
;  automaticamente fechado.
;
;     No final, a caneta ficar sobre o ltimo ponto da sequncia.
;
(define pinta  
  (lambda (seq-pontos) 
    (if graf-cor-temp?
	((flip-solid-polygon graf-vp) 
	 (list->posn-list seq-pontos '(0 0))
	 (make-posn 0 0))
	((draw-solid-polygon graf-vp) 
	 (list->posn-list seq-pontos '(0 0))
	 (make-posn 0 0)
	 graf-cor-caneta))         
    ;                                      
    ;  e deixa caneta no ltimo ponto
    ;
    (let ((ultimo (car (reverse seq-pontos)))) 
      (set-car! graf-posi (car ultimo))
      (set-car! (cdr graf-posi) (cadr ultimo)))))
;
; pinta oval inscrita num rectngulo
;
; O rectngulo  definido por uma lista de 2 vrtices,
; especificados pela seguinte ordem: 
;             superior-esquerdo e inferior-direiro
;
;         No final, a caneta mantm a posio de partida.
;
(define pinta-oval        
  (lambda (dois-vertices)  
    (let* ((v1 (car dois-vertices)) 
	   (v2 (cadr dois-vertices))
	   (v1x (car v1))
	   (v1y (cadr v1))
	   (larg (- (car v2) (car v1)))
	   (alt (- (cadr v1) (cadr v2))))
      (if graf-cor-temp?
          ((flip-solid-ellipse graf-vp) (pt->posn v1 '(0 0)) larg alt)  
          
          ((draw-solid-ellipse graf-vp) (pt->posn v1 '(0 0)) larg alt graf-cor-caneta)))))
;
; Como desenha, 
; mas os pontos da sequncia so definidos em modo relativo.
;
(define desenha-rel  
  (lambda (seq-pontos)      
    (let ((aux
	   (lambda (elem)
	     (begin 
	       (set-car! graf-posi (+ (car elem)
                                      (car graf-posi)))
	       (set-car! (cdr graf-posi) (+ (cadr elem)
                                            (cadr graf-posi)))
	       (list (car graf-posi)
		     (cadr graf-posi))))))
      
      (desenha (map aux 
                    seq-pontos)))))


;
; Como desenha-oval, 
; mas os vrtices so ambos definidos relativamente
;  posio corrente da caneta.
;
(define desenha-oval-rel          
  (lambda (dois-vertices)   
    (let ((corrente-x (car graf-posi)) 
          (corrente-y (cadr graf-posi))  
          (ver1 (car dois-vertices))
          (ver2 (cadr dois-vertices)))
      (desenha-oval (list
		     (list (+ corrente-x (car ver1))
			   (+ corrente-y (cadr ver1)))
		     (list (+ corrente-x (car ver2))
			   (+ corrente-y (cadr ver2))))))))

;
; Como pinta, 
; mas os pontos da sequncia so definidos em modo relativo.
;
(define pinta-rel 
  (lambda (seq-pontos)   
    (let ((aux
	   (lambda (elem)
	     (begin 
	       (set-car! graf-posi (+ (car elem)
                                      (car graf-posi)))
	       (set-car! (cdr graf-posi) (+ (cadr elem)
                                            (cadr graf-posi)))
	       (list (car graf-posi)
		     (cadr graf-posi))))))
      (pinta (map aux 
                  ;(cons '(0 0) seq-pontos))))))
                  seq-pontos)))))

;
; Como pinta-oval, 
; mas os vrtices so ambos definidos relativamente
;  posio corrente da caneta.
;
(define pinta-oval-rel         
  (lambda (dois-vertices)  
    (let ((corrente-x (car graf-posi))  ; ao ponto corrente
          (corrente-y (cadr graf-posi))  ; e ponto corrente mantm-se
          (ver1 (car dois-vertices))
          (ver2 (cadr dois-vertices)))
      (pinta-oval (list
                   (list (+ corrente-x (car ver1))
                         (+ corrente-y (cadr ver1)))
                   (list (+ corrente-x (car ver2))
                         (+ corrente-y (cadr ver2))))))))
;
; Desenha um ponto
;     No final, a caneta ficar sobre o ponto.
;
(define ponto    ; desenha ponto p
  (lambda (p)   ; e deixa caneta em p
    (if graf-cor-temp?
        ((flip-pixel graf-vp) (make-posn (car p) 
                                         (- graf-altura (cadr p))) graf-cor-caneta)
        ((draw-pixel graf-vp) (make-posn (car p) 
                                         (- graf-altura (cadr p))) graf-cor-caneta))
    ;
    (set-car! graf-posi (car p))
    (set-car! (cdr graf-posi) (cadr p))))

; Como ponto,
; mas a coordenada ponto  definida relativamente 
;  posio da caneta
;
(define ponto-rel  
  (lambda (p)       
    (ponto (list (+ (car p)
		    (car graf-posi))
		 (+ (cadr p)
		    (cadr graf-posi))))))
; desenha texto,
; a partir da posio corrente da caneta
;
(define desenha-txt 
  (lambda (texto)
    (if graf-cor-temp?
        ((flip-string graf-vp) 
         (make-posn (car graf-posi)
                    (- graf-altura
                       (cadr graf-posi)))
         texto)
        
        ((draw-string graf-vp) 
         (make-posn (car graf-posi)
                    (- graf-altura
                       (cadr graf-posi)))
         texto graf-cor-caneta))     
    ;
    ;  e deixa caneta na parte final do texto...
    ;
    (let ((larg-alt ((get-string-size graf-vp) texto)))  
      (set-car! graf-posi (+ (car larg-alt)
                             (car graf-posi))))))
;
; Devolve a posio actual da caneta
;
(define caneta 
  (lambda ()
    (list (car graf-posi)
	  (cadr graf-posi))))

(define cursor
  (lambda ()
    (let* ((click-rato (get-mouse-click graf-vp))
           (posi-cursor (mouse-click-posn click-rato)))
      (if (left-mouse-click? click-rato)
          (list (posn-x posi-cursor)
                (- graf-altura (posn-y posi-cursor)))
          (cursor)))))

;
; Limpa a janela grfica
;
(define limpa
  (lambda ()
    ((clear-viewport graf-vp))))

;
; ------------------------------
;
; Procedimentos auxiliares
; 
;
; Converte uma lista de pontos numa lista de pontos
; do DrScheme...
;
(define list->posn-list
  (lambda (s-pontos pos)
    (let
	((x-posi (car pos))
	 (y-posi (cadr pos)))
      (let
	  ((aux
	    (lambda (elem)
	      (make-posn
	       (+ (car elem)
		  x-posi)
	       (- graf-altura
		  (+ (cadr elem)
		     y-posi))))))
	(map aux s-pontos)))))
;
; Converte um ponto num ponto do DrScheme
;
(define pt->posn
  (lambda (pt pos)
    (let ((x-posi (car pos))
	  (y-posi (cadr pos)))
      (make-posn
       (+ (car pt)
          x-posi)
       (- graf-altura
          (+ (cadr pt)
             y-posi))))))



; Fim do ficheiro
